# Generate a loop kernel from that cell integral. This will map the UFL IR
# to the loopy IR.
from ufl.algorithms import MultiFunction
# Spread the pymbolic import statements to where they are used.
from pymbolic.primitives import Variable, Subscript, Sum, Product
import loopy
import numpy
import ufl

from dune.perftool.restriction import Restriction

# Define the generators that are used here
from dune.perftool.generation import generator_factory
loopy_iname = generator_factory(item_tags=("loopy", "kernel", "iname"))
loopy_expr_instruction = generator_factory(item_tags=("loopy", "kernel", "instruction", "exprinstruction"), no_deco=True)
loopy_temporary_variable = generator_factory(item_tags=("loopy", "kernel", "temporary"), on_store=lambda n: loopy.TemporaryVariable(n, dtype=numpy.float64), no_deco=True)
loopy_c_instruction = generator_factory(item_tags=("loopy", "kernel", "instruction", "cinstruction"), no_deco=True)
loopy_valuearg = generator_factory(item_tags=("loopy", "kernel", "argument", "valuearg"), on_store=lambda n: loopy.ValueArg(n), no_deco=True)

@generator_factory(item_tags=("loopy", "kernel", "argument", "globalarg"))
def loopy_globalarg(name, shape=loopy.auto):
    if isinstance(shape, str):
        shape = (shape,)
    return loopy.GlobalArg(name, numpy.float64, shape)

@generator_factory(item_tags=("loopy", "kernel", "domain"))
def loopy_domain(iname, shape):
    loopy_valuearg(shape)
    return "{{ [{0}] : 0<={0}<{1} }}".format(iname, shape)

@loopy_iname
def dimension_iname(index):
    from dune.perftool.pdelab import name_index
    from dune.perftool.pdelab.geometry import name_dimension
    iname = name_index(index)
    dimname = name_dimension()
    loopy_domain(iname, dimname)
    return iname

@loopy_iname
def argument_iname(arg):
    # TODO extract the {iname}_n thing by a preamble
    iname = "arg{}".format(chr(ord("i") + arg.number()))
    loopy_domain(iname, iname + "_n")
    return iname

@loopy_iname
def quadrature_iname():
    loopy_domain("q", "q_n")
    return "q"


class UFLVisitor(MultiFunction):
    def __call__(self, o, init_inames):
        # Make this multifunction stateful: Store information similar to uflacs' modified terminal
        self.grad = False
        self.reference_grad = False
        self.index = None
        self.restriction = Restriction.NONE

        # Have a short cut for the bases class call operator
        self.call = lambda *a : MultiFunction.__call__(self, *a)

        # Have a list of arguments that this term depends on.
        self.arguments = []
        self.inames = init_inames

        # Visit the expression. The return value will be a pymbolic expression.
        loopyexpr = self.call(o)

        # The expression is completely processes, an accumulation instruction can be issued!
        # Have some data structures for the analysis of the given arguments.
        accumargs = []

        # visit the given arguments and extract necessary information
        for arg, index, grad, restriction in sorted(self.arguments, key=lambda x: x[0].number()):
            from dune.perftool.pdelab.argument import name_argumentspace
            accumargs.append(name_argumentspace(arg, index, restriction))
            accumargs.append(argument_iname(arg))

        # Explicitly state the dependency on the quadrature iname
        self.inames.append(quadrature_iname())

        # Evaluate the update factor
        from dune.perftool.pdelab.quadrature import name_factor
        loopyexpr = Product((loopyexpr, Variable(name_factor())))
        # TODO: Some unique name mangling on this temporary variable is necessary!
        expr_tv = loopy_temporary_variable("expr")
        loopy_expr_instruction(loopy.ExpressionInstruction(assignee=Variable("expr"), expression=loopyexpr))

        from dune.perftool.pdelab.argument import name_residual
        residual = name_residual()

        loopy_c_instruction(loopy.CInstruction(self.inames, "{}.accumulate({}, {})".format(residual, ", ".join(accumargs), Variable("xyz"))))


    def grad(self, o):
        assert len(o.operands()) == 1

        self.grad = True
        ret = self.call(o.operands()[0])
        self.grad = False
        return ret

    def argument(self, o):
        assert len(o.operands()) == 0

        # Construct a tuple that represents this argument and all its modifications
        self.arguments.append((o, self.index, self.grad, self.restriction))

        # Generate the variable name for the test function
        from dune.perftool.pdelab.argument import name_testfunction
        name = name_testfunction(o, self.index, self.grad, self.restriction)
        index = argument_iname(o)
        self.inames.append(index)
        loopy_globalarg(name)

        return Subscript(Variable(name), (Variable(index),))

    def coefficient(self, o):
        assert len(o.operands()) == 0
        # TODO: Implement coefficient functions
        assert o.count() == 0

        # TODO implement non-trialfunction coefficients!
        from dune.perftool.pdelab.argument import name_trialfunction
        from dune.perftool.pdelab import name_index
        name = name_trialfunction(o, self.index, self.grad, self.restriction)
        loopy_globalarg(name)

        return Variable(name)

    def product(self, o):
        return Product(tuple(self.call(op) for op in o.operands()))

    def multi_index(self, o):
        # I don't think we should ever find a multi index, because its father should take care of it.
        raise ValueError

    def indexed(self, o):
        # TODO currently, we assume single indices
        assert len(o.operands()[1]) == 1

        # We have two types of indices in our expressions:
        #  * those that should result in variable subscripts
        #  * those that should result in function space child selection
        # I think it is correct to assume that Indexed expressions
        # are of the latter type if and only if their first operand is
        # an argument or coefficient.
        if isinstance(o.operands()[0], (ufl.Argument, ufl.Coefficient)):
            assert not self.index
            self.index = o.operands()[1]
            ret = self.call(o.operands()[0])
            self.index = None
            return ret
        else:
            from dune.perftool.pdelab import name_index
            ret = self.call(o.operands()[0])
            index = Variable(name_index(o.operands()[1]))
            if isinstance(ret, Subscript):
                return Subscript(ret.aggregate, ret.index + (index,))
            else:
                return Subscript(ret, (index,))


class TopSumSeparation(MultiFunction):
    """ A multifunction that separates the toplevel sum """
    def __init__(self, visitor=UFLVisitor()):
        MultiFunction.__init__(self)
        self.visitor = visitor
        self.inames = []

    def expr(self, o):
        self.visitor(o, self.inames)
        self.inames = []

    def sum(self, o):
        for op in o.operands():
            self(op)

    def index_sum(self, o):
        # Generate an iname: We do this here and restrict ourselves to shape dim.
        # TODO revisit when implementing general index sums
        if isinstance(self.visitor, UFLVisitor):
            self.inames.append(dimension_iname(o.operands()[1]))
        self(o.operands()[0])


def transform_expression(expr):
    TopSumSeparation()(expr)
